from .experiment_core import Experiment
from .hafting_2008_data import Hafting2008Data
from .sargolini_2006_data import SargoliniDataTrajectory, Sargolini2006Data
from .wernle_2018_data import Wernle2018Data



import os.path
import warnings
from typing import Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.io as sio

import neuralplayground
from neuralplayground.datasets import fetch_data_path
from neuralplayground.experiments.hafting_2008_data import Hafting2008Data
from neuralplayground.utils import get_2D_ratemap


class Wernle2018Data(Hafting2008Data):
    """Data class for https://www.nature.com/articles/s41593-017-0036-6
    The data can be obtained from https://doi.org/10.11582/2017.00023
    """

    def __init__(
        self,
        data_path: str = None,
        recording_index: int = None,
        experiment_name: str = "FullWernleData",
        verbose: bool = False,
    ):
        """Wernle2018Data init, just initializing parent class Hafting2008Data

        Parameters
        ----------
        data_path: str
            if None, fetch the data from the NeuralPlayground data repository,
            else load data from given path
        recording_index: int
            if None, load data from default recording index
        experiment_name: str
            string to identify object in case of multiple instances
        verbose:
            if True, it will print original readme and data structure when initializing this object
        """
        super().__init__(
            data_path=data_path,
            recording_index=recording_index,
            experiment_name=experiment_name,
            verbose=verbose,
        )

    def _find_data_path(self, data_path: str):
        """Fetch data from NeuralPlayground data repository 
        if no data path is supplied by the user"""
        if data_path is None:
            self.data_path = fetch_data_path("wernle_2018")
        else:
            self.data_path = data_path

    def set_animal_data(self, recording_index: int = 0, tolerance: float = 1e-10):
        """Set position and head direction to be used by the Arena Class later"""
        session_data, rev_vars, rat_info = self.get_recording_data(recording_index)

        # Some of the recorded data in the list do not include positional data
        if type(rev_vars) is list:
            _, x, y = (
                session_data["time"],
                session_data["posx"],
                session_data["posy"],
            )
        else:
            warnings.warn("This index does not have position data")
            return
        pass

        # Position from meters to cm
        self.position = np.stack([x, y], axis=1) * 100
        head_direction = np.diff(self.position, axis=0)
        # Compute head direction from position derivative
        head_direction = head_direction / np.sqrt(np.sum(head_direction**2, axis=1) + tolerance)[..., np.newaxis]
        self.head_direction = head_direction

    def _load_data(self):
        """
        This method loads the recording data of the paper generating the following attributes.
        For further details refer to the readme of the original repo for the paper with the variable description.

        Returns (set the following attributes)
        -------
        self.ratemap : ndarray
            (128 x 2) where each element is a 100x100 ratemap from 10 different rats
            The first column are the reatmaps before merging the room, second column after merging the room.
        self.ratemap_dev : ndarray (19 x 2)
            ORIGINAL DESCRIPTION FROM README: contains ratemaps for the first trial in AB (19 cells from 10 rats);
            rows correspond to single cells, column 1: ratemap A¦B, column 2: first trial ratemap AB
        self.pos_A_B : ndarray
            Position of rats before merging rooms
            (19 x 1) where each element is a (T, 4) recording where T are the total number of samples in time
            at a 50Hz frequency, and column 1 is the time in seconds, column 2 is x_position, column 3 is y_position,
            column 4 is speed index (1 if speed is above 5cm/s)
        self.pos_AB : ndarray
            Position of rats after merging rooms
            (19 x 1) where each element is a (T, 4) recording where T are the total number of samples in time
            at a 50Hz frequency, and column 1 is the time in seconds, column 2 is x_position, column 3 is y_position,
            column 4 is speed index (1 if speed is above 5cm/s)
        """

        # Load ratemaps
        self.inner_path = "nn_Data+Code/data/"
        self.ratemap = sio.loadmat(os.path.join(self.data_path, self.inner_path, "Figures_1_2_3/ratemaps.mat"))
        self.ratemap = self.ratemap["ratemaps"]

        # The following is the recording progression of grid cell patterns for 19 different cells
        # over 10 different rats

        self.ratemap_dev = sio.loadmat(os.path.join(self.data_path, self.inner_path, r"Figure 4/ratemapsDevelopment.mat"))[
            "ratemapsDevelopment"
        ]

        self.pos_A_B = sio.loadmat(os.path.join(self.data_path, self.inner_path, r"Figure 4/posA_B.mat"))["posA_B"]

        self.pos_AB = sio.loadmat(os.path.join(self.data_path, self.inner_path, r"Figure 4/posAB.mat"))["posAB"]
        self.spikes_AB = sio.loadmat(os.path.join(self.data_path, self.inner_path, r"Figure 4/spkAB.mat"))["spkAB"]
        # Setting Arena Limits in cm
        self.arena_limits = np.array([[-100, 100], [-100, 100]])

    def _create_dataframe(self):
        """Generate dataframe for easy display and access of data"""
        self.list = []
        rec_index = 0

        # Two types of data available, ratemaps, and full trajectories with spikes
        for i in range(self.ratemap_dev.shape[0]):
            # full trajectories with spikes
            self.list.append(
                {
                    "rec_index": rec_index,
                    "session": i,
                    "recorded_vars": [
                        "time",
                        "posx",
                        "posy",
                        "speed_index",
                        "spikes",
                        "ratemap_dev",
                    ],
                    "before_merge": False,
                }
            )
            rec_index += 1
            self.list.append(
                {
                    "rec_index": rec_index,
                    "session": i,
                    "recorded_vars": [
                        "time",
                        "posx",
                        "posy",
                        "speed_index",
                        "ratemap_dev",
                    ],
                    "before_merge": True,
                }
            )
            rec_index += 1

        for i in range(self.ratemap.shape[0]):
            # Ratemaps before and after merge
            self.list.append(
                {
                    "rec_index": rec_index,
                    "session": i,
                    "recorded_vars": "ratemap",
                    "before_merge": True,
                }
            )
            rec_index += 1
            self.list.append(
                {
                    "rec_index": rec_index,
                    "session": i,
                    "recorded_vars": "ratemap",
                    "before_merge": False,
                }
            )
            rec_index += 1

        self.recording_list = pd.DataFrame(self.list).set_index("rec_index")

    def get_recording_data(self, recording_index: Union[int, list, tuple] = None):
        """Get experimental data for a given recordin index1

        Parameters
        ----------
        recording_index: int, list of ints or tuple of ints
            recording identifier, index in pandas dataframe with listed data

        Returns (If given a list of recording_index, it will return a list of following parameters)
        -------
        sess_data: dict
            Dictionary with recorded raw data from the session of the respective recording index
            Format of this data follows original readme from the authors of the experiments
        rec_vars: list of str
            keys of session_data dict, recorded variables for a given session
        identifiers: dict
            Dictionary with rat_id and session_id of the returned session data
        """
        if type(recording_index) is list or type(recording_index) is tuple:
            data_list = []
            for ind in recording_index:
                session_info = self.recording_list.iloc[ind]
                if type(session_info["recorded_vars"]) is list:
                    sess_index = session_info["session"]
                    if session_info["before_merge"]:
                        sess_data = {
                            "time": self.pos_A_B[sess_index, 0][:, 0],
                            "posx": self.pos_A_B[sess_index, 0][:, 1],
                            "posy": self.pos_A_B[sess_index, 0][:, 2],
                            "speed_index": self.pos_A_B[sess_index, 0][:, 3],
                            "ratemap_dev": self.ratemap_dev[sess_index, 0],
                        }
                    else:
                        sess_data = {
                            "time": self.pos_AB[sess_index, 0][:, 0],
                            "posx": self.pos_AB[sess_index, 0][:, 1],
                            "posy": self.pos_AB[sess_index, 0][:, 2],
                            "speed_index": self.pos_AB[sess_index, 0][:, 3],
                            "spikes": self.spikes_AB[sess_index, 0][:, 0],
                            "ratemap_dev": self.ratemap_dev[sess_index, 1],
                        }
                    rev_vars = list(sess_data.keys()) + [
                        "dev",
                    ]
                    identifiers = {"sess_index": sess_index}
                    data_list.append([sess_data, rev_vars, identifiers])

                elif session_info["recorded_vars"] == "ratemap":
                    sess_index = session_info["session"]
                    rev_vars = [
                        "ratemap",
                    ]
                    identifiers = {"sess_index": sess_index}
                    if session_info["before_merge"]:
                        data_list.append(
                            [
                                {"ratemap": self.ratemap[sess_index, 0]},
                                rev_vars,
                                identifiers,
                            ]
                        )
                    else:
                        data_list.append(
                            [
                                {"ratemap": self.ratemap[sess_index, 1]},
                                rev_vars,
                                identifiers,
                            ]
                        )

            return data_list

        else:
            session_info = self.recording_list.iloc[recording_index]
            if type(session_info["recorded_vars"]) is list:
                sess_index = session_info["session"]
                identifiers = {"sess_index": sess_index}
                if session_info["before_merge"]:
                    sess_data = {
                        "time": self.pos_A_B[sess_index, 0][:, 0],
                        "posx": self.pos_A_B[sess_index, 0][:, 1],
                        "posy": self.pos_A_B[sess_index, 0][:, 2],
                        "speed_index": self.pos_A_B[sess_index, 0][:, 3],
                        "ratemap_dev": self.ratemap_dev[sess_index, 0],
                    }
                else:
                    sess_data = {
                        "time": self.pos_AB[sess_index, 0][:, 0],
                        "posx": self.pos_AB[sess_index, 0][:, 1],
                        "posy": self.pos_AB[sess_index, 0][:, 2],
                        "speed_index": self.pos_AB[sess_index, 0][:, 3],
                        "spikes": self.spikes_AB[sess_index, 0][:, 0],
                        "ratemap_dev": self.ratemap_dev[sess_index, 1],
                    }
                rev_vars = list(sess_data.keys()) + [
                    "dev",
                ]
                return sess_data, rev_vars, identifiers

            elif session_info["recorded_vars"] == "ratemap":
                sess_index = session_info["session"]
                identifiers = {"sess_index": sess_index}
                if session_info["before_merge"]:
                    sess_data = {"ratemap": self.ratemap[sess_index, 0]}
                else:
                    sess_data = {"ratemap": self.ratemap[sess_index, 1]}
                rev_vars = "ratemap"
                return sess_data, rev_vars, identifiers

    def plot_recording_tetr(
        self,
        recording_index: Union[int, tuple, list] = None,
        save_path: Union[str, tuple, list] = None,
        ax: Union[mpl.axes.Axes, tuple, list] = None,
        tetrode_id: Union[str, tuple, list] = None,
        bin_size: float = 2.0,
    ):
        """Plot tetrode ratemap from spike data for a given recording index or a list of recording index.
        If given a list or tuple as argument, all arguments must be list, tuple, or None.

        Parameters
        ----------
        recording_index: int, tuple of ints, list of ints
            recording index to plot spike ratemap, if list or tuple, it will recursively call this function
            to make a plot per recording index. If this argument is list or tuple, the rest of variables must
            be list or tuple with their respective types, or keep the default None value.
        save_path: str, list of str, tuple of str
            saving path of the generated figure, if None, no figure is saved
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots), list of ax, or tuple of ax
            axis or list of axis from subplot from matplotlib where the ratemap will be plotted.
            If None, ax is generated with default options.
        tetrode_id: str, list of str, or tuple of str
            tetrode id in the corresponding session
        bin_size: float
            bin size to discretize space when computing ratemap

        Returns
        -------
        h: ndarray (nybins, nxbins)
            Number of spikes falling on each bin through the recorded session, nybins number of bins in y axis,
            nxbins number of bins in x axis
        binx: ndarray (nxbins +1,)
            bin limits of the ratemap on the x axis
        biny: ndarray (nybins +1,)
            bin limits of the ratemap on the y axis
        (when using list pr tuple as argument, this function return a list or tuple of the variables listed above)
        """
        # Recursive call of this function in case of list or tuple
        if type(recording_index) is list or type(recording_index) is tuple:
            axis_list = []
            for i, ind in enumerate(recording_index):
                # Checking if rest of variables are default or list values
                if save_path is not None:
                    save_path_i = save_path[i]
                else:
                    save_path_i = None
                if ax is not None:
                    ax_i = ax[0]
                else:
                    ax_i = None
                if tetrode_id is not None:
                    tetrode_id_i = tetrode_id[i]
                else:
                    tetrode_id_i = None
                ind_axis = self.plot_recording_tetr(ind, save_path=save_path_i, ax=ax_i, tetrode_id=tetrode_id_i)
                axis_list.append(ind_axis)
            return axis_list

        # Generate axis in case ax is None
        if ax is None:
            f, ax = plt.subplots(1, 1, figsize=(10, 8))

        # Recall recorded data
        session_data, rev_vars, rat_info = self.get_recording_data(recording_index)

        arena_width = self.arena_limits[0, 1] - self.arena_limits[0, 0]
        arena_depth = self.arena_limits[1, 1] - self.arena_limits[1, 0]

        # Check of the session has spikes to compute ratemap
        # Plot pre-computed ratemap otherwise
        if type(rev_vars) is list and "spikes" in rev_vars:
            time_array, test_spikes, x, y = (
                session_data["time"],
                session_data["spikes"],
                session_data["posx"],
                session_data["posy"],
            )

            h, binx, biny = get_2D_ratemap(
                time_array,
                test_spikes,
                x,
                y,
                x_size=int(arena_width / bin_size),
                y_size=int(arena_depth / bin_size),
                filter_result=True,
            )

        elif type(rev_vars) is list and "spikes" not in rev_vars:
            warnings.warn("No spike data pre merging")
            return
        else:
            h = session_data["ratemap"]
            binx = np.linspace(self.arena_limits[0, 0], self.arena_limits[0, 1], num=h.shape[1])
            biny = np.linspace(self.arena_limits[1, 0], self.arena_limits[1, 1], num=h.shape[0])

        # Adding merging status to plot title
        merged = self.recording_list.iloc[recording_index]["before_merge"]
        merged_mssg = "merged" if not merged else "before_merge"

        sess_index = self.recording_list.iloc[recording_index]["session"]
        # Use auxiliary function to make the plot
        self._make_tetrode_plot(h, ax, "sess_index_" + str(sess_index) + "_" + merged_mssg, save_path)
        # Save if save_path is not None
        if save_path is None:
            pass
        else:
            plt.savefig(save_path, bbox_inches="tight")
        # Return ratemap values, x bin limits and y bin limits
        return h, binx, biny

    def plot_trajectory(
        self,
        recording_index: Union[int, tuple, list] = None,
        save_path: Union[str, tuple, list] = None,
        ax: Union[mpl.axes.Axes, tuple, list] = None,
        plot_every: int = 20,
    ):
        """Plot animal trajectory from a given recording index, corresponding to a recording session

        Parameters
        ----------
        recording_index: int, tuple of ints, list of ints
            recording index to plot spike ratemap, if list or tuple, it will recursively call this function
            to make a plot per recording index. If this argument is list or tuple, the rest of variables must
            be list or tuple with their respective types, or keep the default None value.
        save_path: str, list of str, tuple of str
            saving path of the generated figure, if None, no figure is saved
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots), list of ax, or tuple of ax
            axis or list of axis from subplot from matplotlib where the ratemap will be plotted.
            If None, ax is generated with default options.
        plot_every: int
            time steps skipped to make the plot to reduce cluttering

        Returns
        -------
        x: ndarray (n_samples,)
            x position throughout recording of the given session
        y: ndarray (n_samples,)
            y position throughout recording of the given session
        time_array: ndarray (n_samples,)
            array with the timestamps in seconds per position of the given session

        """
        if type(recording_index) is list or type(recording_index) is tuple:
            axis_list = []
            for i, ind in enumerate(recording_index):
                # Checking if rest of variables are default or list values
                if save_path is not None:
                    save_path_i = save_path[i]
                else:
                    save_path_i = None
                if ax is not None:
                    ax_i = ax[0]
                else:
                    ax_i = None
                ind_axis = self.plot_trajectory(ind, save_path=save_path_i, ax=ax_i, plot_every=plot_every)
                axis_list.append(ind_axis)
            return axis_list

        # Generate axis in case ax is None
        if ax is None:
            f, ax = plt.subplots(1, 1, figsize=(10, 8))

        session_data, rev_vars, identifiers = self.get_recording_data(recording_index)

        self.arena_limits[0, 1] - self.arena_limits[0, 0]
        self.arena_limits[1, 1] - self.arena_limits[1, 0]

        if type(rev_vars) is list:
            time_array, x, y = (
                session_data["time"],
                session_data["posx"],
                session_data["posy"],
            )
        else:
            warnings.warn("This index does not have position data")
            return

        # Helper function to format the trajectory plot
        ax = self._make_trajectory_plot(x, y, ax, plot_every)
        if save_path is None:
            pass
        else:
            plt.savefig(save_path, bbox_inches="tight")
        return x, y, time_array

    def plot_merging_comparison(self, session_index: Union[int, list, tuple]):
        """Plot ratemaps before and after merging for a given session index

        Parameters
        ----------
        session_index: int, list of ints, tuple of ints
            session id to identify ratemap from dataset
        Returns
        -------
        ratemaps_before: list
            List of ratemaps before the merging
        ratemaps_after: list
            List of ratempas after the merging
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots)
            plot axes from matplotlib for further formating
        """
        if type(session_index) is list or type(session_index) is tuple:
            n_cells = len(session_index)
        else:
            n_cells = 1
            session_index = [
                session_index,
            ]
        f, ax = plt.subplots(n_cells, 2, figsize=(8, 5 * n_cells))
        ratemaps_before = []
        ratemaps_after = []
        for i in range(n_cells):
            ax[i, 0].imshow(self.ratemap[session_index[i], 0], cmap="jet")
            ax[i, 1].imshow(self.ratemap[session_index[i], 1], cmap="jet")
            ratemaps_before.append(self.ratemap[session_index[i], 0])
            ratemaps_after.append(self.ratemap[session_index[i], 1])
            ax[i, 0].axhline(y=50, color="white")
            ax[i, 1].axhline(y=50, color="white", linestyle="--")
        ax[0, 0].set_title("Before merging")
        ax[0, 1].set_title("After merging")
        return ratemaps_before, ratemaps_after, ax

    def plot_development(
        self,
        n_cells: int = 3,
        time_interval: Union[tuple, list] = (1.0, 2.0),
        merged: bool = False,
        plot_every: int = 10,
    ):
        """Prototype function for evolution of ratemaps through time

        Parameters
        ----------
        n_cells: int
            Number of cells to plot, from 0 to 19
        time_interval: 2D-tuple or list with floats
            Time inverval from experiment, if merged true, then time inverval after merging
        merged: bool
            If True, make plot for data after merging
        plot_every: int
            time steps skipped to make the plot to reduce cluttering

        Returns
        -------
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots)
            Modified axis where the trajectory is plotted
        """
        if merged:
            pos = self.pos_AB[:n_cells, 0]
        else:
            pos = self.pos_A_B[:n_cells, 0]

        f, ax = plt.subplots(1, n_cells, figsize=(4 * n_cells, 3))
        for cell in range(n_cells):
            pos_i = pos[cell]
            init_sample = int(pos_i.shape[0] * (time_interval[0] * 60.0) / (np.amax(pos_i[:, 0])))
            finish_sample = int(pos_i.shape[0] * (time_interval[1] * 60.0) / (np.amax(pos_i[:, 0])))
            finish_sample - init_sample
            cmap = mpl.cm.get_cmap("plasma")
            norm = plt.Normalize(init_sample, finish_sample)
            aux_x, aux_y = [], []
            prev_x, prev_y = pos_i[init_sample, 1], pos_i[finish_sample, 2]
            for sample_i in range(init_sample, finish_sample, plot_every):
                x, y = pos_i[sample_i, 1], pos_i[sample_i, 2]
                aux_x.append(x)
                aux_y.append(y)
                ax[cell].plot(
                    (prev_x, x),
                    (prev_y, y),
                    "-",
                    color=cmap(norm(sample_i)),
                    alpha=0.4,
                    lw=0.5,
                )
                prev_x = x
                prev_y = y
            ax[cell].set_ylim((-100, 100))
            ax[cell].set_xlim((-100, 100))
            sc = ax[cell].scatter(
                aux_x,
                aux_y,
                c=np.arange(init_sample, finish_sample, plot_every),
                vmin=init_sample,
                vmax=finish_sample,
                cmap="plasma",
                alpha=0.4,
                s=2,
            )
            ax[cell].axhline(y=0, color="black")
            ticks = np.linspace(init_sample, finish_sample, num=10)
            tickslabels = np.round(np.linspace(init_sample / 50 / 60, finish_sample / 50 / 60, num=10), 1)
            cbar = plt.colorbar(sc, ax=ax[cell], ticks=ticks)
            cbar.ax.set_yticklabels(tickslabels, fontsize=8)
            cbar.ax.set_ylabel("Time [min]", rotation=270, labelpad=12)
        return ax

    def get_recorded_session(self, recording_index=None):
        # Not used, override to avoid issues
        return [None, None, None]


if __name__ == "__main__":
    data = Wernle2018Data()
    data.show_data()
    data.plot_cell_comparison(session_index=(125, 126))



import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter


def check_crossing_wall(
    pre_state: np.ndarray,
    new_state: np.ndarray,
    wall: np.ndarray,
    wall_closenes: float = 1e-5,
    tolerance: float = 1e-9,
):
    """

    Parameters
    ----------
    pre_state : (2,) 2d-ndarray
        2d position of pre-movement
    new_state : (2,) 2d-ndarray
        2d position of post-movement
    wall : (2, 2) ndarray
        [[x1, y1], [x2, y2]] where (x1, y1) is on limit of the wall, (x2, y2) second limit of the wall
    wall_closenes : float
        how close the agent is allowed to be from the wall
    tolerance: float
        Small constant to avoid inverting a singular matrix

    Returns
    -------
    new_state: (2,) 2d-ndarray
        corrected new state. If it is not crossing the wall, then the new_state stays the same, if the state cross the
        wall, new_state will be corrected to a valid place without crossing the wall
    cross_wall: bool
        True if the change in state cross a wall
    """

    # Check if the line of the wall and the line between the states cross
    A = np.stack([np.diff(wall, axis=0)[0, :], -new_state + pre_state], axis=1)
    b = pre_state - wall[0, :]
    try:
        intersection = np.linalg.inv(A) @ b
    except Exception:
        intersection = np.linalg.inv(A + np.identity(A.shape[0]) * tolerance) @ b
    smaller_than_one = intersection <= 1
    larger_than_zero = intersection >= 0

    # If condition is true, then the points cross the wall
    cross_wall = np.alltrue(np.logical_and(smaller_than_one, larger_than_zero))
    if cross_wall:
        new_state = (intersection[-1] - wall_closenes) * (new_state - pre_state) + pre_state

    return new_state, cross_wall


def create_circular_wall(center: np.ndarray, radius: float, n_walls: int = 100):
    """Generate a circular wall by discretizing the circle into many walls.

    Parameters
    ----------
    center: ndarray (2,)
        Center of the circular wall
    radius: float
        Radius of the circular wall
    n_walls: int
        Number of walls used to discretize the circle

    Returns
    -------
    list_of_segments: list of walls
        n_walls that creates a circle
    """

    d_angle = 2 * np.pi / n_walls
    list_of_segments = []
    for i in range(n_walls):
        init_point = np.array([radius * np.cos(d_angle * i), radius * np.sin(d_angle * i)]) + center
        end_point = np.array([radius * np.cos(d_angle * (i + 1)), radius * np.sin(d_angle * (i + 1))]) + center
        wall = np.stack([init_point, end_point])
        list_of_segments.append(wall)
    return list_of_segments


def inheritors(klass):
    """Given a class, returns child classes (for future implementation of comparison board)"""
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses


def clean_data(data, keep_headers=False):
    """For a dictionary with positional data, this function replace nans by interpolated positions
    Used in data readers in the experiment classes

    Parameters
    ----------
    data: dict
        dictionary with values that might contain arrays with nans
    keep_headers: bool
        it will remove unwanted keys from the returned dictionary that are residuals of
        loading .mat files
        (sounds a bit cryptic, please check _load_data in experiment classes)

    Returns
    -------
    aux_dict: dict
        Modified dictionary with some keys removed and nans interpolated
    """
    aux_dict = {}
    for key, val in data.items():
        if isinstance(val, bytes) or isinstance(val, str) or key == "__globals__":
            if keep_headers:
                aux_dict[key] = val
            continue
        else:
            if not np.isnan(val).any():
                aux_dict[key] = val
            else:
                # Interpolate nans
                x_range = np.linspace(0, 1, num=len(val))
                nan_indexes = np.logical_not(np.isnan(val))[:, 0]
                clean_x = x_range[nan_indexes]
                clean_val = np.array(val)[nan_indexes, 0]
                # print(clean_x.shape, clean_val.shape)
                f = interp1d(clean_x, clean_val, kind="cubic", fill_value="extrapolate")
                aux_dict[key] = f(x_range)[..., np.newaxis]
    return aux_dict


def get_2D_ratemap(
    time_array: np.ndarray,
    spikes: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    x_size: float = 50,
    y_size: float = 50,
    filter_result: bool = False,
):
    """
    Parameters
    ----------
    time_array: ndarray (n_samples,)
        array with the timestamps in seconds per position of the given session
    spikes: ndarray (n_spikes,)
        spike times in seconds of the given session
    x: ndarray (n_samples,)
        x position throughout recording of the given session
    y: ndarray (n_samples,)
        y position throughout recording of the given session
    x_size: float
        bin size in x dim
    y_size: float
        bin size in y dim
    filter_result: bool
        If true, it will smooth the ratemap using a gaussian filter

    Returns
    -------
    h: ndarray (nybins, nxbins)
        Number of spikes falling on each bin through the recorded session, nybins number of bins in y axis,
        nxbins number of bins in x axis
    binx: ndarray (nxbins +1,)
        bin limits of the ratemap on the x axis
    biny: ndarray (nybins +1,)
        bin limits of the ratemap on the y axis
    """
    x_spikes, y_spikes = [], []

    # Find x, y position of each spike using spike times
    for s in spikes:
        array_pos = np.argmin(np.abs(time_array - s))
        x_spikes.append(x[array_pos])
        y_spikes.append(y[array_pos])
    x_spikes = np.array(x_spikes)
    y_spikes = np.array(y_spikes)
    h, binx, biny = np.histogram2d(x_spikes, y_spikes, bins=(x_size, y_size))
    # Gaussian filter
    if filter_result:
        h = gaussian_filter(h, sigma=2)
    # Transpose to change row col coordinates to positions
    return h.T, binx, biny


class OnlineRateMap(object):
    def __init__(
        self,
        spikes,
        position,
        size=(100, 100),
        x_range=(-100, 100),
        y_range=(-100, 100),
    ):
        """
        Creates a ratemap where it is possible to distinguish firing rate and traversed position at the same time

        Parameters
        ----------
        spikes : ndarray
            (n_spikes,) shaped array with times of spikes for a single cell
        position : ndarray
            (timestamps, 3) shaped array with timestamp time in seconds,
             position in x-dim on column 2 and y-dim on column 2 for each timestamp
        size : tuple
            (2,) shaped tupple with the size of the desired ratemap
        x_range : tuple
            (2,) shaped tuple with limits on x-dim
        y_range
            (2,) shaped tuple with limits on y-dim
        """
        self.ratemap = np.empty(shape=size)
        self.ratemap[:] = np.nan
        self.size = size
        self.x_range = x_range
        self.y_range = y_range
        self.x_pos_bins = np.linspace(x_range[0], x_range[1], num=size[0] + 1)
        self.y_pos_bins = np.linspace(y_range[0], y_range[1], num=size[1] + 1)
        self.spikes = spikes
        self.position = position
        self.last_t_end = 0
        self.last_t_init = 0

    def get_ratemap(self, t_end, t_init=0, interp_factor=10):
        """
        Computes the ratemap using given spike train (self.spikes) and position (self.position)

        Parameters
        ----------
        t : float
            maximum time of recording to consider for the ratemap
        t_init : float
            starting time of recording to consider for the ratemap, default set to 0
        Returns
        -------
        ratemap : ndarray
            updated ratemap for times between t_init and t
        """

        # Get spikes and position within the range
        pos = self.position[
            np.logical_and(self.position[:, 0] >= t_init, self.position[:, 0] < t_end),
            :,
        ]

        # Interpolate position for smooth plot
        if interp_factor != 1:
            t = pos[:, 0]
            x = pos[:, 1]
            y = pos[:, 2]
            t_inter = np.linspace(np.amin(t), np.amax(t), num=len(t) * interp_factor, endpoint=False)
            fx = interp1d(t, x)
            fy = interp1d(t, y)
            x_inter = fx(t_inter)
            y_inter = fy(t_inter)
            pos = np.stack([t_inter, x_inter, y_inter], axis=1)

        spk = self.spikes[np.logical_and(self.spikes >= t_init, self.spikes < t_end)]

        h_spk, bin_spk = np.histogram(spk, bins=np.linspace(t_init, t_end, num=pos.shape[0] + 1))

        # Convert position to indexes in the ratemap
        x_index = (pos[:, 1] - self.x_range[0]) / (self.x_range[1] - self.x_range[0]) * self.size[1]
        y_index = (pos[:, 2] - self.y_range[0]) / (self.y_range[1] - self.y_range[0]) * self.size[0]
        x_index, y_index = x_index.astype(int), y_index.astype(int)

        # update ratemap pixels
        for pos_i in range(len(x_index)):
            t = pos[pos_i, 0]
            ratemap_val = self.ratemap[y_index[pos_i], x_index[pos_i]]
            if np.isnan(ratemap_val):
                self.ratemap[y_index[pos_i], x_index[pos_i]] = 0
            self.ratemap[y_index[pos_i], x_index[pos_i]] += h_spk[pos_i]

        smooth_ratemap = self.get_smooth_ratemap()

        self.last_t_init = t_init
        self.last_t_end = t_end
        return smooth_ratemap

    def update_ratemap(self, dt, interp_factor=1):
        t_init = self.last_t_init
        t_end = self.last_t_end + dt
        pos = self.position[
            np.logical_and(self.position[:, 0] >= t_init, self.position[:, 0] < t_end),
            :,
        ]

        # Interpolate position for smooth plot
        if interp_factor != 1:
            t = pos[:, 0]
            x = pos[:, 1]
            y = pos[:, 2]
            t_inter = np.linspace(np.amin(t), np.amax(t), num=len(t) * interp_factor, endpoint=False)
            fx = interp1d(t, x)
            fy = interp1d(t, y)
            x_inter = fx(t_inter)
            y_inter = fy(t_inter)
            pos = np.stack([t_inter, x_inter, y_inter], axis=1)

        spk = self.spikes[np.logical_and(self.spikes >= t_init, self.spikes < t_end)]

        h_spk, bin_spk = np.histogram(spk, bins=np.linspace(t_init, t_end, num=pos.shape[0] + 1))

        # Convert position to indexes in the ratemap
        x_index = (pos[:, 1] - self.x_range[0]) / (self.x_range[1] - self.x_range[0]) * self.size[1]
        y_index = (pos[:, 2] - self.y_range[0]) / (self.y_range[1] - self.y_range[0]) * self.size[0]
        x_index, y_index = x_index.astype(int), y_index.astype(int)

        # update ratemap pixels
        for pos_i in range(len(x_index)):
            t = pos[pos_i, 0]
            ratemap_val = self.ratemap[y_index[pos_i], x_index[pos_i]]
            if np.isnan(ratemap_val):
                self.ratemap[y_index[pos_i], x_index[pos_i]] = 0
            self.ratemap[y_index[pos_i], x_index[pos_i]] += h_spk[pos_i]

        smooth_ratemap = self.get_smooth_ratemap()

        self.last_t_init = t_init
        self.last_t_end = t_end
        return smooth_ratemap

    def get_smooth_ratemap(self):
        nan_indexes = np.isnan(self.ratemap)
        aux_ratemap = np.copy(self.ratemap)
        aux_ratemap[nan_indexes] = 0
        filtered_ratemap = gaussian_filter(aux_ratemap, 3.5)
        self.nan = np.nan
        filtered_ratemap[nan_indexes] = self.nan
        return filtered_ratemap



import glob
import os.path

import numpy as np
import scipy.io as sio

import neuralplayground
from neuralplayground.datasets import fetch_data_path
from neuralplayground.experiments import Experiment, Hafting2008Data
from neuralplayground.utils import clean_data


class SargoliniDataTrajectory(Experiment):
    """Data class for sargolini et al. 2006. https://www.science.org/doi/10.1126/science.1125572
    The data can be obtained from https://archive.norstore.no/pages/public/datasetDetail.jsf?id=8F6BE356-3277-475C-87B1-C7A977632DA7
    This class only consider animal trajectory pre-processed by the authors.
    """

    def __init__(
        self,
        experiment_name: str = "Sargolini_2006_Data",
        data_path: str = None,
        **kwargs,
    ):
        """SargoliniData Class Init

        Parameters
        ----------
        experiment_name: str
            string to identify object in case of multiple instances
        data_path: str
            if None, fetch the data from the NeuralPlayground data repository,
            else load data from given path
        """
        self.experiment_name = experiment_name
        if data_path is None:
            # Set data_path to the data directory within the package
            self.data_path = fetch_data_path("sargolini_2006")
        else:
            self.data_path = data_path
        # Sort the data in data_path
        (
            self.arena_limits,
            self.position,
            self.head_direction,
            self.time_span,
        ) = self._get_sargolini_data()

    def _get_sargolini_data(self, tolerance: float = 1e-10):
        """Load and concatenate animal trajectories from data pre-processed by the authors

        Parameters
        ----------
        tolerance: float
            Small constant to avoid dividing by zero when estimating head direction

        Returns
        -------
        arena_limits: ndarray (2, 2)
            first row x limits of the arena, second row y limits of the arena, in cm
        position: ndarray (n, 2)
            first column is x pos in cm, second column is y pos in cm, n is the number of sampled positions
        head_direction: ndarray (n-1, 2)
            directional vector with estimated head direction from position derivative
        time_span: ndarray (n, )
            time stamp per measurement assuming 50Hz of sampling rate
        """
        # Arena limits in cm, sampling rate in Hz, both from the original experiment
        arena_limits = np.array([[-50, 50], [-50, 50]])
        self.sampling_rate = 50

        # Listing files with trajectories
        filenames_x = os.path.join(self.data_path, "sargolini_x_pos_")
        filenames_y = os.path.join(self.data_path, "sargolini_y_pos_")

        # Filing array from files
        x_position = np.array([])
        y_position = np.array([])
        for i in range(61):
            aux_x = np.load(filenames_x + str(i) + ".npy")
            aux_y = np.load(filenames_y + str(i) + ".npy")
            x_position = np.concatenate([x_position, aux_x])
            y_position = np.concatenate([y_position, aux_y])

        position = np.stack([x_position, y_position], axis=1) * 100  # Convert to cm, originally in meters
        head_direction = np.diff(position, axis=0)  # Head direction from derivative of position
        head_direction = head_direction / np.sqrt(np.sum(head_direction**2, axis=1) + tolerance)[..., np.newaxis]
        time_span = np.arange(head_direction.shape[0]) * (1 / self.sampling_rate)
        return arena_limits, position, head_direction, time_span


class Sargolini2006Data(Hafting2008Data):
    """Data class for sargolini et al. 2006. https://www.science.org/doi/10.1126/science.1125572
    The data can be obtained from https://archive.norstore.no/pages/public/datasetDetail.jsf?id=8F6BE356-3277-475C-87B1-C7A977632DA7
    This class only consider animal raw animal trajectories and neural recordings.
    """

    def __init__(
        self,
        data_path: str = None,
        recording_index: int = None,
        experiment_name: str = "FullSargoliniData",
        verbose: bool = False,
    ):
        """Sargolini2006Data init, just initializing parent class Hafting2008Data

        Parameters
        ----------
        data_path: str
            if None, fetch the data from the NeuralPlayground data repository,
            else load data from given path
        recording_index: int
            if None, load data from default recording index
        experiment_name: str
            string to identify object in case of multiple instances
        verbose:
            if True, it will print original readme and data structure when initializing this object
        """
        super().__init__(
            data_path=data_path,
            recording_index=recording_index,
            experiment_name=experiment_name,
            verbose=verbose,
        )

    def _find_data_path(self, data_path: str):
        """Fetch data from NeuralPlayground data repository 
        if no data path is supplied by the user"""
        if data_path is None:
            self.data_path = fetch_data_path("sargolini_2006") + "raw_data_sample/"
        else:
            self.data_path = data_path + "raw_data_sample/"

    def _load_data(self):
        """Parse data according to specific data format
        if you are a user check the notebook examples"""
        self.best_recording_index = 0  # Nice session recording as default
        # Arena limits from the experimental setting, first row x limits, second row y limits, in cm
        self.arena_limits = np.array([[-50.0, 50.0], [-50.0, 50.0]])
        data_path_list = glob.glob(self.data_path + "*.mat")
        mice_ids = np.unique([dp.split("/")[-1][:5] for dp in data_path_list])
        # Initialize data dictionary, later handled by this object itself (so don't worry about this)
        self.data_per_animal = {}
        for m_id in mice_ids:
            m_paths_list = glob.glob(self.data_path + m_id + "*.mat")
            sessions = np.unique([dp.split("/")[-1].split("-")[1][:8] for dp in m_paths_list]).astype(str)
            self.data_per_animal[m_id] = {}
            for sess in sessions:
                s_paths_list = glob.glob(self.data_path + m_id + "-" + sess + "*.mat")
                cell_ids = np.unique([dp.split("/")[-1].split(".")[-2][-4:] for dp in s_paths_list]).astype(str)
                self.data_per_animal[m_id][sess] = {}
                for cell_id in cell_ids:
                    if cell_id == "_POS":
                        session_info = "position"
                    elif cell_id in ["_EEG", "_EGF"]:
                        session_info = cell_id[1:]
                    else:
                        session_info = cell_id
                    r_path = glob.glob(self.data_path + m_id + "-" + sess + "*" + cell_id + "*.mat")
                    # Interpolate to replace NaNs and stuff
                    cleaned_data = clean_data(sio.loadmat(r_path[0]))
                    if cell_id != "_POS" and cell_id not in ["_EEG", "_EGF"]:
                        try:
                            self.data_per_animal[m_id][sess][session_info] = cleaned_data["cellTS"]
                        except Exception:
                            pass
                    else:
                        self.data_per_animal[m_id][sess][session_info] = cleaned_data

    def get_tetrode_data(self, session_data: str = None, tetrode_id: str = None):
        """Return time stamp, position and spikes for a given session and tetrode

        Parameters
        ----------
        session_data: str
            if None, the session used corresponds to the default recording index
        tetrode_id:
            tetrode id in the corresponding session

        Returns
        -------
        time_array: ndarray (n_samples,)
            array with the timestamps in seconds per position of the given session
        test_spikes: ndarray (n_spikes,)
            spike times in seconds of the given session
        x: ndarray (n_samples,)
            x position throughout recording of the given session
        y: ndarray (n_samples,)
            y position throughout recording of the given session
        """
        if session_data is None:
            session_data, rev_vars, rat_info = self.get_recording_data(recording_index=0)
            tetrode_id = self._find_tetrode(rev_vars)
        position_data = session_data["position"]
        x1, y1 = position_data["posx"][:, 0], position_data["posy"][:, 0]
        x2, y2 = x1, y1
        # Selecting positional data
        x = np.clip(x2, a_min=self.arena_limits[0, 0], a_max=self.arena_limits[0, 1])
        y = np.clip(y2, a_min=self.arena_limits[1, 0], a_max=self.arena_limits[1, 1])
        time_array = position_data["post"][:]
        tetrode_data = session_data[tetrode_id]
        test_spikes = tetrode_data[:, 0]
        time_array = time_array[:, 0]

        return time_array, test_spikes, x, y


if __name__ == "__main__":
    # print("initializing hafting")
    # data = FullHaftingData(verbose=True)
    # print("plotting_tragectory")
    # data.plot_trajectory(2)
    # print("plotting_recording")
    # data.plot_recording_tetr(2)
    # plt.show()

    print("initializing sargolini")
    # data = FullSargoliniData(verbose=True)
    # print("plotting_tragectory")
    # data.plot_trajectory(2)
    # print("plotting_recording")
    # data.plot_recording_tetr(2)
    # plt.show()



import cv2
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from gymnasium.spaces import Box
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

from neuralplayground.arenas.arena_core import Environment
from neuralplayground.utils import check_crossing_wall


class Simple2D(Environment):
    """
    Methods (Some in addition to Environment class)
    ----------
    __init__(self, environment_name="2DEnv", **env_kwargs):
        Initialise the class
    reset(self):
        Reset the environment variables
    step(self, action):
        Increment the global step count of the agent in the environment and moves
        the agent in a random direction with a fixed step size
    plot_trajectory(self, history_data=None, ax=None):
        Plot the Trajectory of the agent in the environment. In addition to environment class.
    _create_default_walls(self):
        Generates outer border of the 2D environment based on the arena limits

    Attributes (Some in addition to the Environment class)
    ----------
    state: ndarray
        Contains the x, y coordinate of the position and head direction of the agent (will be further developed)
        head_direction: ndarray
                Contains the x and y Coordinates of the position
        position: ndarray
                Contains the x and y Coordinates of the position
    history: list of dicts
        Saved history over simulation steps (action, state, new_state, reward, global_steps)
    global_steps: int
        Counter of the number of steps in the environment
    room_width: int
        Size of the environment in the x direction
    room_depth: int
        Size of the environment in the y direction
    metadata: dict
        Dictionary containing the metadata
    observation: ndarray
        Fully observable environment, make_observation returns the state
        Array of the observation of the agent in the environment (Could be modified as the environments are evolves)
    agent_step_size: float
         Size of the step when executing movement, agent_step_size*global_steps will give
         a measure of the total distance traversed by the agent
    """

    def __init__(self, environment_name: str = "2DEnv", **env_kwargs):
        """Initialise the class

        Parameters
        ----------
        env_kwargs: dict
        Dictionary with parameters of the experiment of the children class
            time_step_size: float
                time_step_size * global_steps will give a measure of the time in the experimental setting
            agent_step_size: float
                Step size used when the action is a direction in x,y coordinate (normalize false in step())
                Agent_step_size * global_step_number will give a measure of the distance in the experimental setting
            arena_x_limits: float
                Size of the environment in the x direction
            arena_y_limits: float
                Size of the environment in the y direction

        environment_name: str
            Name of the specific instantiation of the Simple2D class

        """
        super().__init__(environment_name, **env_kwargs)
        self.metadata = {"env_kwargs": env_kwargs}
        self.arena_x_limits, self.arena_y_limits = (
            env_kwargs["arena_x_limits"],
            env_kwargs["arena_y_limits"],
        )
        self.arena_limits = np.array(
            [
                [self.arena_x_limits[0], self.arena_x_limits[1]],
                [self.arena_y_limits[0], self.arena_y_limits[1]],
            ]
        )
        self.room_width = np.diff(self.arena_x_limits)[0]
        self.room_depth = np.diff(self.arena_y_limits)[0]
        self.observation_space = Box(
            low=np.array([self.arena_x_limits[0], self.arena_y_limits[0]]),
            high=np.array([self.arena_x_limits[1], self.arena_y_limits[1]]),
            dtype=np.float64,
        )
        # Action are not bounded in this environment since the action is normalized in the step function
        # And can't cross the walls of the environment
        self.action_space = Box(
            low=np.array((-np.inf, -np.inf)),
            high=np.array((np.inf, np.inf)),
            dtype=np.float64,
        )
        self.agent_step_size = env_kwargs["agent_step_size"]
        self.state_dims_labels = ["x_pos", "y_pos"]

        self._create_default_walls()
        self._create_custom_walls()
        self.wall_list = self.default_walls + self.custom_walls
        self.state_dims_labels = ["x_pos", "y_pos"]
        self.reset()

    def _create_default_walls(self):
        """Generate walls to limit the arena based on the limits given in kwargs when initializing the object.
        Each wall is presented by a matrix
            [[xi, yi],
             [xf, yf]]
        where xi and yi are x y coordinates of one limit of the wall, and xf and yf are coordinates of the other limit.
        Walls are added to default_walls list, to then merge it with custom ones.
        See notebook with custom arena examples.
        """
        self.default_walls = []
        self.default_walls.append(
            np.array(
                [
                    [self.arena_limits[0, 0], self.arena_limits[1, 0]],
                    [self.arena_limits[0, 0], self.arena_limits[1, 1]],
                ]
            )
        )
        self.default_walls.append(
            np.array(
                [
                    [self.arena_limits[0, 1], self.arena_limits[1, 0]],
                    [self.arena_limits[0, 1], self.arena_limits[1, 1]],
                ]
            )
        )
        self.default_walls.append(
            np.array(
                [
                    [self.arena_limits[0, 0], self.arena_limits[1, 0]],
                    [self.arena_limits[0, 1], self.arena_limits[1, 0]],
                ]
            )
        )
        self.default_walls.append(
            np.array(
                [
                    [self.arena_limits[0, 0], self.arena_limits[1, 1]],
                    [self.arena_limits[0, 1], self.arena_limits[1, 1]],
                ]
            )
        )

    def _create_custom_walls(self):
        """Custom walls method. In this case is empty since the environment is a simple square room
        Override this method to generate more walls, see jupyter notebook with examples
        """
        self.custom_walls = []

    def reset(self, random_state: bool = False, custom_state: np.ndarray = None):
        """Reset the environment variables

        Parameters
        ----------
        random_state: bool
            If True, sample a new position uniformly within the arena, use default otherwise
        custom_state: np.ndarray
            If given, use this array to set the initial state

        Returns
        ----------
        observation: ndarray
            Because this is a fully observable environment, make_observation returns the state of the environment
            Array of the observation of the agent in the environment (Could be modified as the environments are evolves)

        self.state: ndarray (2,)
            Vector of the x and y coordinate of the position of the animal in the environment
        """
        self.global_steps = 0
        self.global_time = 0
        self.history = []
        if random_state:
            self.state = [
                np.random.uniform(low=self.arena_limits[0, 0], high=self.arena_limits[0, 1]),
                np.random.uniform(low=self.arena_limits[1, 0], high=self.arena_limits[1, 1]),
            ]
        else:
            self.state = [0, 0]
        self.state = np.array(self.state)

        if custom_state is not None:
            self.state = np.array(custom_state)
        # Fully observable environment, make_observation returns the state
        observation = self.make_observation()
        return observation, self.state

    def step(self, action: np.ndarray, normalize_step: bool = False):
        """Runs the environment dynamics. Increasing global counters.
        Given some action, return observation, new state and reward.

        Parameters
        ----------
        action: ndarray (2,)
            Array containing the action of the agent, in this case the delta_x and detla_y increment to position
        normalize_step: bool
            If true, the action is normalized to have unit size, then scaled by the agent step size

        Returns
        -------
        reward: float
            The reward that the animal receives in this state
        new_state: ndarray
            Update the state with the updated vector of coordinate x and y of position and head directions respectively
        observation: ndarray
            Array of the observation of the agent in the environment
        """
        if normalize_step:
            action = action / np.linalg.norm(action)
            new_state = self.state + self.agent_step_size * action
        else:
            new_state = self.state + action
        new_state, valid_action = self.validate_action(self.state, action, new_state)
        reward = self.reward_function(action, self.state)  # If you get reward, it should be coded here
        transition = {
            "action": action,
            "state": self.state,
            "next_state": new_state,
            "reward": reward,
            "step": self.global_steps,
        }
        self.history.append(transition)
        self.state = new_state
        observation = self.make_observation()
        self._increase_global_step()
        return observation, new_state, reward

    def validate_action(self, pre_state, action, new_state):
        """Check if the new state is crossing any walls in the arena.

        Parameters
        ----------
        pre_state : (2,) 2d-ndarray
            2d position of pre-movement
        new_state : (2,) 2d-ndarray
            2d position of post-movement

        Returns
        -------
        new_state: (2,) 2d-ndarray
            corrected new state. If it is not crossing the wall, then the new_state stays the same, if the state cross the
            wall, new_state will be corrected to a valid place without crossing the wall
        crossed_wall: bool
            True if the change in state crossed a wall and was corrected
        """
        crossed_wall = False
        for wall in self.wall_list:
            new_state, crossed = check_crossing_wall(pre_state=pre_state, new_state=new_state, wall=wall)
            crossed_wall = crossed or crossed_wall
        return new_state, crossed_wall

    def plot_trajectory(
        self,
        history_data: list = None,
        ax=None,
        return_figure: bool = False,
        save_path: str = None,
        plot_every: int = 1,
    ):
        """Plot the Trajectory of the agent in the environment

        Parameters
        ----------
        history_data: list of interactions
            if None, use history data saved as attribute of the arena, use custom otherwise
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots)
            axis from subplot from matplotlib where the trajectory will be plotted.
        return_figure: bool
            If true, it will return the figure variable generated to make the plot
        save_path: str, list of str, tuple of str
            saving path of the generated figure, if None, no figure is saved

        Returns
        -------
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots)
            Modified axis where the trajectory is plotted
        f: matplotlib.figure
            if return_figure parameters is True
        """

        # Use or not saved history
        if history_data is None:
            history_data = self.history

        # Generate Figure
        if ax is None:
            f, ax = plt.subplots(1, 1, figsize=(8, 6))

        # Draw walls
        for wall in self.default_walls:
            ax.plot(wall[:, 0], wall[:, 1], "C3", lw=3)

        # Draw custom walls
        for wall in self.custom_walls:
            ax.plot(wall[:, 0], wall[:, 1], "C0", lw=3)

        # Making the trajectory plot roughly square to show structure of the arena better
        lower_lim, upper_lim = np.amin(self.arena_limits), np.amax(self.arena_limits)
        ax.set_xlim([lower_lim, upper_lim])
        ax.set_ylim([lower_lim, upper_lim])

        # Make plot of positions
        if len(history_data) != 0:
            state_history = [s["state"] for s in history_data]
            next_state_history = [s["next_state"] for s in history_data]
            state_history[0]
            next_state_history[-1]

            cmap = mpl.cm.get_cmap("plasma")
            norm = plt.Normalize(0, len(state_history))

            aux_x = []
            aux_y = []
            for i, s in enumerate(state_history):
                if i % plot_every == 0:
                    if i + plot_every >= len(state_history):
                        break
                    x_ = [s[0], state_history[i + plot_every][0]]
                    y_ = [s[1], state_history[i + plot_every][1]]
                    aux_x.append(s[0])
                    aux_y.append(s[1])
                    sc = ax.plot(x_, y_, "-", color=cmap(norm(i)), alpha=0.6)

            sc = ax.scatter(
                aux_x,
                aux_y,
                c=np.arange(len(aux_x)),
                vmin=0,
                vmax=len(aux_x),
                cmap="plasma",
                alpha=0.6,
                s=0.5,
            )
            cbar = plt.colorbar(sc, ax=ax, ticks=[0, len(state_history)])
            cbar.ax.set_ylabel("N steps", rotation=270, fontsize=16)
            cbar.ax.set_yticklabels([0, len(state_history)], fontsize=16)

        if save_path is not None:
            plt.savefig(save_path, bbox_inches="tight")

        if return_figure:
            return ax, f
        else:
            return ax

    def render(self, history_length=30):
        """Render the environment live through iterations"""
        f, ax = plt.subplots(1, 1, figsize=(8, 6))
        canvas = FigureCanvas(f)
        history = self.history[-history_length:]
        ax = self.plot_trajectory(history_data=history, ax=ax)
        canvas.draw()
        image = np.frombuffer(canvas.tostring_rgb(), dtype="uint8")
        image = image.reshape(f.canvas.get_width_height()[::-1] + (3,))
        print(image.shape)
        cv2.imshow("2D_env", image)
        cv2.waitKey(10)



import numpy as np

from ..experiments.wernle_2018_data import Wernle2018Data
from .hafting_2008 import Hafting2008
from .simple2d import Simple2D


class Wernle2018(Hafting2008):
    """Arena resembling Wernle2018 experimental setting

    Methods (In addition to Hafting 2008)
    ----------
    set_room(self, room_id: str):
        Place the agent in the right room configuration depending the amount of exploration time
    def step(self, action: np.ndarray, normalize_step: bool = False, skip_every: int = 10):
        Move the agents based on the action and room configuration

    Attributes (In addition to Hafting 2008)
    ----------
    merge_time: float
        Time in minutes to remove the middle wall in the experiment, merging the two rooms
    switch_time: float
        Time in minutes to change the agent from one room to the other one
    """

    def __init__(
        self,
        use_behavioral_data: bool = False,
        data_path: str = None,
        recording_index: int = None,
        environment_name: str = "Wernle2018",
        verbose: bool = False,
        experiment_class=Wernle2018Data,
        merge_time: float = 100,
        switch_time: float = 50,
        **env_kwargs,
    ):
        """

        Parameters
        ----------
        use_behavioral_data: bool
            If True, then uses the animal trajectories recorded in Wernle 2018
        data_path: str
            if None, fetch the data from the NeuralPlayground data repository,
            else load data from given path
        recording_index: int
            if None, load data from default recording index of corresponding experiment class
        environment_name: str
            Name of the specific instantiation of the Wernle 2018 class
        verbose: bool
            Set to True to show the information of the class
        experiment_class:
            Experiment class to be initialized
        merge_time: float
            Time in minutes to remove the middle wall in the experiment, merging the two rooms
        switch_time: float
            Time in minutes to change the agent from one room to the other one
        env_kwargs: dict (keys below)
            time_step_size: float
                Time step size in second
        """
        super().__init__(
            use_behavioral_data,
            data_path,
            recording_index,
            environment_name,
            verbose,
            experiment_class,
            **env_kwargs,
        )
        self.time_step_size = env_kwargs["time_step_size"]
        self.merge_time = int((merge_time * 60) / self.time_step_size)
        self.switch_time = int((switch_time * 60) / self.time_step_size)

        self.AB_id = "AB"
        self.A_id = "A"
        self.B_id = "B"

    def set_room(self, room_id: str):
        """Place the agent in the right room configuration depending the amount of exploration time

        Parameters
        ----------
        room_id: str
            Either "A", "B" or "AB". If "A" or "B" it will locate the agent in one of the rooms with a wall in between.
            If "AB", it will place the agent in a random position and remove the wall in between
        """
        if room_id == self.A_id:
            # Take the agent to room A
            self.state = [
                np.random.uniform(low=self.arena_limits[0, 0], high=self.arena_limits[0, 1]),
                np.random.uniform(low=self.arena_limits[1, 0], high=0),
            ]
            # Add wall in between to separete rooms
            self.custom_walls = [np.array([[-100, 0], [100, 0]])]
            # Update walls in the environment
            self.wall_list = self.default_walls + self.custom_walls
        elif room_id == self.B_id:
            # Take the agent to room B
            self.state = [
                np.random.uniform(low=self.arena_limits[0, 0], high=self.arena_limits[0, 1]),
                np.random.uniform(low=0, high=self.arena_limits[1, 1]),
            ]
            # Add wall in between to separete rooms
            self.custom_walls = [np.array([[-100, 0], [100, 0]])]
            # Update walls in the environment
            self.wall_list = self.default_walls + self.custom_walls
        elif room_id == self.AB_id:
            if len(self.custom_walls) != 0:
                # Remove wall in between to merge rooms
                self.custom_walls = []
                # Update walls in the environment
                self.wall_list = self.default_walls + self.custom_walls

    def _create_custom_walls(self):
        """Add wall in between when initializing environment as default"""
        self.custom_walls = [np.array([[-100, 0], [100, 0]])]

    def step(self, action: np.ndarray, normalize_step: bool = False, skip_every: int = 10):
        """Set the right room configuration, then call default step function"""
        if self.use_behavioral_data:
            return super().step(action, normalize_step, skip_every)
        if self.global_steps == 0:
            self.set_room("A")
        elif self.global_steps == self.merge_time:
            self.set_room("AB")
        elif self.global_steps == self.switch_time:
            self.set_room("B")
        return super().step(action, normalize_step, skip_every)


class MergingRoom(Simple2D):
    """Arena resembling Wernle2018 experimental setting BUT WITH GENERIC ROOM DIMENSIONS

    Methods (In addition to Simple2D)
    ----------
    set_room(self, room_id: str):
        Place the agent in the right room configuration depending the amount of exploration time
    def step(self, action: np.ndarray, normalize_step: bool = False, skip_every: int = 10):
        Move the agents based on the action and room configuration

    Attributes (In addition to Simple2D)
    ----------
    merge_time: float
        Time in minutes to remove the middle wall in the experiment, merging the two rooms
    switch_time: float
        Time in minutes to change the agent from one room to the other one
    """

    def __init__(
        self,
        environment_name: str = "Wernle2018",
        merge_time: float = 100,
        switch_time: float = 50,
        **env_kwargs,
    ):
        """

        Parameters
        ----------
        use_behavioral_data: bool
            If True, then uses the animal trajectories recorded in Wernle 2018
        data_path: str
            if None, fetch the data from the NeuralPlayground data repository,
            else load data from given path
        recording_index: int
            if None, load data from default recording index of corresponding experiment class
        environment_name: str
            Name of the specific instantiation of the Wernle 2018 class
        verbose: bool
            Set to True to show the information of the class
        experiment_class:
            Experiment class to be initialized
        merge_time: float
            Time in minutes to remove the middle wall in the experiment, merging the two rooms
        switch_time: float
            Time in minutes to change the agent from one room to the other one
        env_kwargs: dict (keys below)
            time_step_size: float
                Time step size in second
        """

        super().__init__(environment_name, **env_kwargs)

        self.time_step_size = env_kwargs["time_step_size"]
        self.merge_time = int((merge_time * 60) / self.time_step_size)
        self.switch_time = int((switch_time * 60) / self.time_step_size)

        self.AB_id = "AB"
        self.A_id = "A"
        self.B_id = "B"

    def set_room(self, room_id: str):
        """Place the agent in the right room configuration depending the amount of exploration time

        Parameters
        ----------
        room_id: str
            Either "A", "B" or "AB". If "A" or "B" it will locate the agent in one of the rooms with a wall in between.
            If "AB", it will place the agent in a random position and remove the wall in between
        """
        if room_id == self.A_id:
            # Take the agent to room A
            self.state = [
                np.random.uniform(low=self.arena_limits[0, 0], high=self.arena_limits[0, 1]),
                np.random.uniform(low=self.arena_limits[1, 0], high=0),
            ]
            # Add wall in between to separete rooms
            self.custom_walls = [np.array([[self.arena_limits[0, 0], 0], [self.arena_limits[0, 1], 0]])]
            # Update walls in the environment
            self.wall_list = self.default_walls + self.custom_walls
        elif room_id == self.B_id:
            # Take the agent to room B
            self.state = [
                np.random.uniform(low=self.arena_limits[0, 0], high=self.arena_limits[0, 1]),
                np.random.uniform(low=0, high=self.arena_limits[1, 1]),
            ]
            # Add wall in between to separete rooms
            self.custom_walls = [np.array([[self.arena_limits[0, 0], 0], [self.arena_limits[0, 1], 0]])]
            # Update walls in the environment
            self.wall_list = self.default_walls + self.custom_walls
        elif room_id == self.AB_id:
            if len(self.custom_walls) != 0:
                # Remove wall in between to merge rooms
                self.custom_walls = []
                # Update walls in the environment
                self.wall_list = self.default_walls + self.custom_walls

    def _create_custom_walls(self):
        """Add wall in between when initializing environment as default"""
        self.custom_walls = [np.array([[self.arena_limits[0, 0], 0], [self.arena_limits[0, 1], 0]])]

    def step(self, action: np.ndarray, normalize_step: bool = False):
        """Set the right room configuration, then call default step function"""
        if self.global_steps == 0:
            self.set_room("A")
        elif self.global_steps == self.merge_time:
            self.set_room("AB")
        elif self.global_steps == self.switch_time:
            self.set_room("B")
        return super().step(action, normalize_step)



import glob
import os.path
from typing import Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.io as sio
from IPython.display import display

from neuralplayground.datasets import fetch_data_path
from neuralplayground.utils import clean_data, get_2D_ratemap

from .experiment_core import Experiment


class Hafting2008Data(Experiment):
    """Data class for Hafting et al. 2008. https://www.nature.com/articles/nature06957
    The data can be obtained from https://archive.norstore.no/pages/public/datasetDetail.jsf?id=C43035A4-5CC5-44F2-B207-126922523FD9
    This class only consider animal raw animal trajectories and neural recordings
    This class is also used for Sargolini2006Data due to its similar data structure
    """

    def __init__(
        self,
        data_path: str = None,
        recording_index: int = None,
        experiment_name: str = "FullHaftingData",
        verbose: bool = False,
    ):
        """Hafting2008Data Init

        Parameters
        ----------
        data_path: str
            if None, fetch the data from the NeuralPlayground data repository,
            else load data from given path
        recording_index: int
            if None, load data from default recording index
        experiment_name: str
            string to identify object in case of multiple instances
        verbose:
            if True, it will print original readme and data structure when initializing this object
        """
        self.experiment_name = experiment_name
        self._find_data_path(data_path)
        self._load_data()
        self._create_dataframe()
        self.rat_id, self.sess, self.rec_vars = self.get_recorded_session(recording_index)
        self.set_animal_data()
        if verbose:
            self.show_readme()
            self.show_data()

    def set_animal_data(self, recording_index: int = 0, tolerance: float = 1e-10):
        """Set position and head direction to be used by the Arena Class later"""
        session_data, rev_vars, rat_info = self.get_recording_data(recording_index)
        tetrode_id = self._find_tetrode(rev_vars)
        time_array, test_spikes, x, y = self.get_tetrode_data(session_data, tetrode_id)

        self.position = np.stack([x, y], axis=1)
        head_direction = np.diff(self.position, axis=0)
        # Compute head direction from position derivative
        head_direction = head_direction / np.sqrt(np.sum(head_direction**2, axis=1) + tolerance)[..., np.newaxis]
        self.head_direction = head_direction

    def _find_data_path(self, data_path: str):
        """Fetch data from NeuralPlayground data repository 
        if no data path is supplied by the user"""
        if data_path is None:
            self.data_path = fetch_data_path("hafting_2008")
        else:
            self.data_path = data_path

    def _load_data(self):
        """Parse data according to specific data format
        if you are a user check the notebook examples"""
        self.best_recording_index = 4  # Nice session recording as default
        # Arena limits from the experimental setting, first row x limits, second row y limits, in cm
        self.arena_limits = np.array([[-200, 200], [-20, 20]])
        data_path_list = glob.glob(self.data_path + "*.mat")
        mice_ids = np.unique([dp.split("/")[-1][:5] for dp in data_path_list])
        # Initialize data dictionary, later handled by this object itself (so don't worry about this)
        self.data_per_animal = {}
        for m_id in mice_ids:
            m_paths_list = glob.glob(self.data_path + m_id + "*.mat")
            sessions = np.unique([dp.split("/")[-1].split("-")[1][:8] for dp in m_paths_list]).astype(str)
            self.data_per_animal[m_id] = {}
            for sess in sessions:
                s_paths_list = glob.glob(self.data_path + m_id + "-" + sess + "*.mat")
                cell_ids = np.unique([dp.split("/")[-1].split(".")[-2][-4:] for dp in s_paths_list]).astype(str)
                self.data_per_animal[m_id][sess] = {}
                for cell_id in cell_ids:
                    if cell_id == "_POS":
                        session_info = "position"
                    elif "EG" in cell_id:
                        session_info = cell_id[1:]
                    else:
                        session_info = cell_id

                    r_path = glob.glob(self.data_path + m_id + "-" + sess + "*" + cell_id + "*.mat")
                    # Interpolate to replace NaNs and stuff
                    cleaned_data = clean_data(sio.loadmat(r_path[0]))
                    self.data_per_animal[m_id][sess][session_info] = cleaned_data

    def _create_dataframe(self):
        """Generate dataframe for easy display and access of data"""
        self.list = []
        idx = 0
        for rat_id, rat_sess in self.data_per_animal.items():
            for sess, recorded_vars in rat_sess.items():
                self.list.append(
                    {
                        "rec_index": idx,
                        "rat_id": rat_id,
                        "session": sess,
                        "recorded_vars": list(recorded_vars.keys()),
                    }
                )
                idx += 1
        self.recording_list = pd.DataFrame(self.list).set_index("rec_index")

    def show_data(self, full_dataframe: bool = False):
        """Print of available data recorded in the experiment

        Parameters
        ----------
        full_dataframe: bool
            if True, it will show all available data, a small sample otherwise

        Returns
        -------
        recording_list: Pandas dataframe
            List of available data, columns with rat_id, recording session and recorded variables
        """
        print("Dataframe with recordings")
        if full_dataframe:
            pd.set_option("display.max_rows", None)
            pd.set_option("display.max_columns", None)
        display(self.recording_list)
        return self.recording_list

    def show_readme(self):
        """Print original readme of the dataset"""
        readme_path = glob.glob(self.data_path + "readme" + "*.txt")[0]
        with open(readme_path, "r") as fin:
            print(fin.read())

    def get_recorded_session(self, recording_index: int = None):
        """Get identifiers to sort the experimental data

        Parameters
        ----------
        recording_index: int
            recording identifier, index in pandas dataframe with listed data

        Returns
        -------
        rat_id: str
            rat identifier from experiment
        sess: str
            recording session identifier from experiment
        recorded_vars: list of str
            Variables recorded from a given session
        """
        if recording_index is None:
            recording_index = self.best_recording_index
        list_item = self.recording_list.iloc[recording_index]
        rat_id, sess, recorded_vars = (
            list_item["rat_id"],
            list_item["session"],
            list_item["recorded_vars"],
        )
        return rat_id, sess, recorded_vars

    def get_recording_data(self, recording_index: int = None):
        """Get experimental data for a given recordin index

        Parameters
        ----------
        recording_index: int
            recording identifier, index in pandas dataframe with listed data

        Returns
        -------
        session_data: dict
            Dictionary with recorded raw data from the session of the respective recording index
            Format of this data follows original readme from the authors of the experiments
        rec_vars: list of str
            keys of session_data dict, recorded variables for a given session
        identifiers: dict
            Dictionary with rat_id and session_id of the returned session data
        """
        if recording_index is None:
            recording_index = self.best_recording_index
        if type(recording_index) is list or type(recording_index) is tuple:
            data_list = []
            for ind in recording_index:
                rat_id, sess, rec_vars = self.get_recorded_session(ind)
                session_data = self.data_per_animal[rat_id][sess]
                data_list.append([session_data, rec_vars, {"rat_id": rat_id, "session": sess}])
            return data_list
        else:
            rat_id, sess, rec_vars = self.get_recorded_session(recording_index)
            session_data = self.data_per_animal[rat_id][sess]
            identifiers = {"rat_id": rat_id, "sess": sess}
            return session_data, rec_vars, identifiers

    def _find_tetrode(self, rev_vars: list):
        """Static function to find tetrode id in a multiple tetrode recording session

        Parameters
        ----------
        rev_vars: list of strings
            recorded variables for a given session, tetrode id within it

        Returns
        -------
        tetrode_id: str
            found first tetrode id in the recorded variable list
        """
        tetrode_id = next(
            var_name for var_name in rev_vars if (var_name != "position") and (("t" in var_name) or ("T" in var_name))
        )
        return tetrode_id

    def get_tetrode_data(self, session_data: str = None, tetrode_id: str = None):
        """Return time stamp, position and spikes for a given session and tetrode

        Parameters
        ----------
        session_data: str
            if None, the session used corresponds to the default recording index
        tetrode_id:
            tetrode id in the corresponding session

        Returns
        -------
        time_array: ndarray (n_samples,)
            array with the timestamps in seconds per position of the given session
        test_spikes: ndarray (n_spikes,)
            spike times in seconds of the given session
        x: ndarray (n_samples,)
            x position throughout recording of the given session
        y: ndarray (n_samples,)
            y position throughout recording of the given session
        """
        # Use default recording index when session is not given
        if session_data is None:
            session_data, rev_vars, rat_info = self.get_recording_data(recording_index=self.best_recording_index)
            tetrode_id = self._find_tetrode(rev_vars)

        position_data = session_data["position"]
        x1, y1 = position_data["posx"][:, 0], position_data["posy"][:, 0]
        x2, y2 = x1, y1
        # Selecting positional data
        x = np.clip(x2, a_min=self.arena_limits[0, 0], a_max=self.arena_limits[0, 1])
        y = np.clip(y2, a_min=self.arena_limits[1, 0], a_max=self.arena_limits[1, 1])
        time_array = position_data["post"][:]
        tetrode_data = session_data[tetrode_id]
        test_spikes = tetrode_data["ts"][:,]
        test_spikes = test_spikes[:, 0]
        time_array = time_array[:, 0]

        return time_array, test_spikes, x, y

    def plot_recording_tetr(
        self,
        recording_index: Union[int, tuple, list] = None,
        save_path: Union[str, tuple, list] = None,
        ax: Union[mpl.axes.Axes, tuple, list] = None,
        tetrode_id: Union[str, tuple, list] = None,
        bin_size: float = 2.0,
    ):
        """Plot tetrode ratemap from spike data for a given recording index or a list of recording index.
        If given a list or tuple as argument, all arguments must be list, tuple, or None.

        Parameters
        ----------
        recording_index: int, tuple of ints, list of ints
            recording index to plot spike ratemap, if list or tuple, it will recursively call this function
            to make a plot per recording index. If this argument is list or tuple, the rest of variables must
            be list or tuple with their respective types, or keep the default None value.
        save_path: str, list of str, tuple of str
            saving path of the generated figure, if None, no figure is saved
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots), list of ax, or tuple of ax
            axis or list of axis from subplot from matplotlib where the ratemap will be plotted.
            If None, ax is generated with default options.
        tetrode_id: str, list of str, or tuple of str
            tetrode id in the corresponding session
        bin_size: float
            bin size to discretize space when computing ratemap

        Returns
        -------
        h: ndarray (nybins, nxbins)
            Number of spikes falling on each bin through the recorded session, nybins number of bins in y axis,
            nxbins number of bins in x axis
        binx: ndarray (nxbins +1,)
            bin limits of the ratemap on the x axis
        biny: ndarray (nybins +1,)
            bin limits of the ratemap on the y axis
        (when using list pr tuple as argument, this function return a list or tuple of the variables listed above)
        """

        # Recursive call of this function in case of list or tuple
        if type(recording_index) is list or type(recording_index) is tuple:
            axis_list = []
            for i, ind in enumerate(recording_index):
                # Checking if rest of variables are default or list values
                if save_path is not None:
                    save_path_i = save_path[i]
                else:
                    save_path_i = None
                if ax is not None:
                    ax_i = ax[0]
                else:
                    ax_i = None
                if tetrode_id is not None:
                    tetrode_id_i = tetrode_id[i]
                else:
                    tetrode_id_i = None
                ind_axis = self.plot_recording_tetr(ind, save_path=save_path_i, ax=ax_i, tetrode_id=tetrode_id_i)
                axis_list.append(ind_axis)
            return axis_list

        # Generate axis in case ax is None
        if ax is None:
            f, ax = plt.subplots(1, 1, figsize=(10, 8))

        # Recall recorded data
        session_data, rev_vars, rat_info = self.get_recording_data(recording_index)
        if tetrode_id is None:
            tetrode_id = self._find_tetrode(rev_vars)

        arena_width = self.arena_limits[0, 1] - self.arena_limits[0, 0]
        arena_depth = self.arena_limits[1, 1] - self.arena_limits[1, 0]

        # Recall spike data
        time_array, test_spikes, x, y = self.get_tetrode_data(session_data, tetrode_id)

        # Compute ratemap matrices from data
        h, binx, biny = get_2D_ratemap(
            time_array,
            test_spikes,
            x,
            y,
            x_size=int(arena_width / bin_size),
            y_size=int(arena_depth / bin_size),
            filter_result=True,
        )

        # Use auxiliary function to make the plot
        self._make_tetrode_plot(h, ax, tetrode_id, save_path)
        # Return ratemap values, x bin limits and y bin limits
        return h, binx, biny

    def _make_tetrode_plot(self, h, ax, title, save_path):
        """plot function with formating of ratemap plot

        Parameters
        ----------
        h: ndarray (nybins, nxbins)
            Number of spikes falling on each bin through the recorded session, nybins number of bins in y axis,
            nxbins number of bins in x axis
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots)
            axis from subplot from matplotlib where the ratemap will be plotted.
        title: str
            plot title, tetrode id by default when called
        save_path: str, list of str, tuple of str
            saving path of the generated figure, if None, no figure is saved

        Returns
        -------
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots)
            Modified axis where ratemap is plotted
        """

        # Formating ratemap plot
        sc = ax.imshow(h, cmap="jet")
        cbar = plt.colorbar(sc, ax=ax, ticks=[np.min(h), np.max(h)], orientation="horizontal")
        cbar.ax.set_xlabel("Firing rate", fontsize=12)
        cbar.ax.set_xticklabels([np.round(np.min(h)), np.round(np.max(h))], fontsize=12)
        ax.set_title(title)

        ax.set_ylabel("width", fontsize=16)
        ax.set_xlabel("depth", fontsize=16)
        ax.grid(False)

        ax.set_xticks([])
        ax.set_yticks([])

        # Save if save_path is not None
        if save_path is None:
            return ax
        else:
            plt.savefig(save_path, bbox_inches="tight")
            return ax

    def plot_trajectory(
        self,
        recording_index: Union[int, tuple, list] = None,
        save_path: Union[str, tuple, list] = None,
        ax: Union[mpl.axes.Axes, tuple, list] = None,
        plot_every: int = 20,
    ):
        """Plot animal trajectory from a given recording index, corresponding to a recording session

        Parameters
        ----------
        recording_index: int, tuple of ints, list of ints
            recording index to plot spike ratemap, if list or tuple, it will recursively call this function
            to make a plot per recording index. If this argument is list or tuple, the rest of variables must
            be list or tuple with their respective types, or keep the default None value.
        save_path: str, list of str, tuple of str
            saving path of the generated figure, if None, no figure is saved
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots), list of ax, or tuple of ax
            axis or list of axis from subplot from matplotlib where the ratemap will be plotted.
            If None, ax is generated with default options.
        plot_every: int
            time steps skipped to make the plot to reduce cluttering

        Returns
        -------
        x: ndarray (n_samples,)
            x position throughout recording of the given session
        y: ndarray (n_samples,)
            y position throughout recording of the given session
        time_array: ndarray (n_samples,)
            array with the timestamps in seconds per position of the given session

        """
        if type(recording_index) is list or type(recording_index) is tuple:
            axis_list = []
            for i, ind in enumerate(recording_index):
                # Checking if rest of variables are default or list values
                if save_path is not None:
                    save_path_i = save_path[i]
                else:
                    save_path_i = None
                if ax is not None:
                    ax_i = ax[0]
                else:
                    ax_i = None
                ind_axis = self.plot_trajectory(ind, save_path=save_path_i, ax=ax_i, plot_every=plot_every)
                axis_list.append(ind_axis)
            return axis_list

        # Generate axis in case ax is None
        if ax is None:
            f, ax = plt.subplots(1, 1, figsize=(8, 6))

        session_data, rev_vars, rat_info = self.get_recording_data(recording_index)
        tetrode_id = self._find_tetrode(rev_vars)

        time_array, test_spikes, x, y = self.get_tetrode_data(session_data, tetrode_id)
        # Helper function to format the trajectory plot
        self._make_trajectory_plot(x, y, ax, plot_every)
        # Save if save_path is not None
        if save_path is None:
            pass
        else:
            plt.savefig(save_path, bbox_inches="tight")
        return x, y, time_array

    def _make_trajectory_plot(self, x, y, ax, plot_every, fontsize=24):
        """

        Parameters
        ----------
        x: ndarray (n_samples,)
            x position throughout recording of the given session
        y: ndarray (n_samples,)
            y position throughout recording of the given session
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots)
            axis from subplot from matplotlib where the ratemap will be plotted.
        plot_every: int
            time steps skipped to make the plot to reduce cluttering
        fontsize: int
            fontsize of labels in the plot

        Returns
        -------
        ax: mpl.axes._subplots.AxesSubplot (matplotlib axis from subplots)
            Modified axis where the trajectory is plotted
        """

        # Plotting borders of the arena
        ax.plot(
            [self.arena_limits[0, 0], self.arena_limits[0, 0]],
            [self.arena_limits[1, 0], self.arena_limits[1, 1]],
            "C3",
            lw=3,
        )
        ax.plot(
            [self.arena_limits[0, 1], self.arena_limits[0, 1]],
            [self.arena_limits[1, 0], self.arena_limits[1, 1]],
            "C3",
            lw=3,
        )
        ax.plot(
            [self.arena_limits[0, 0], self.arena_limits[0, 1]],
            [self.arena_limits[1, 1], self.arena_limits[1, 1]],
            "C3",
            lw=3,
        )
        ax.plot(
            [self.arena_limits[0, 0], self.arena_limits[0, 1]],
            [self.arena_limits[1, 0], self.arena_limits[1, 0]],
            "C3",
            lw=3,
        )

        # Setting colormap of trajectory
        cmap = mpl.cm.get_cmap("plasma")
        norm = plt.Normalize(0, np.size(x))

        aux_x = []
        aux_y = []
        for i in range(len(x)):
            if i % plot_every == 0:
                if i + plot_every >= len(x):
                    break
                x_ = [x[i], x[i + plot_every]]
                y_ = [y[i], y[i + plot_every]]
                aux_x.append(x[i])
                aux_y.append(y[i])
                sc = ax.plot(x_, y_, "-", color=cmap(norm(i)), alpha=0.6)

        # Setting plot labels
        ax.set_xlabel("width", fontsize=fontsize)
        ax.set_ylabel("depth", fontsize=fontsize)
        ax.set_title("position", fontsize=fontsize)
        ax.grid(False)

        cmap = mpl.cm.get_cmap("plasma")
        norm = plt.Normalize(0, np.size(x))
        sc = ax.scatter(
            aux_x,
            aux_y,
            c=np.arange(len(aux_x)),
            vmin=0,
            vmax=len(x),
            cmap="plasma",
            alpha=0.6,
            s=0.1,
        )

        # Setting colorbar to show number of sampled (time steps) recorded
        cbar = plt.colorbar(sc, ax=ax, ticks=[0, len(x)])
        cbar.ax.tick_params(labelsize=fontsize)
        cbar.ax.set_ylabel("N steps", rotation=270, fontsize=fontsize)
        cbar.ax.set_yticklabels([0, len(x)], fontsize=fontsize)
        ax.set_xlim([np.amin([x.min(), y.min()]) - 1.0, np.amax([x.max(), y.max()]) + 1.0])
        ax.set_ylim([np.amin([x.min(), y.min()]) - 1.0, np.amax([x.max(), y.max()]) + 1.0])
        return ax



import copy

from neuralplayground.arenas import Hafting2008
from neuralplayground.experiments import (
    Sargolini2006Data,
    SargoliniDataTrajectory,
)


class Sargolini2006(Hafting2008):
    """Arena resembling Sargolini2006 experimental setting

    Methods
    ----------
    __init__(self, use_behavioral_data: bool = False, data_path: str = None, recording_index: int = None,
                 environment_name: str = "Sargolini2006", verbose: bool = False, experiment_class=Sargolini2006Data,
                 **env_kwargs):
         Initialise the class
    reset(self):
        Reset the environment variables
    step(self, action):
        Increment the global step count of the agent in the environment and updates the position of the agent according
        to the recordings of the specific chosen session

    Attributes
    ----------
    self.state: array
        Contains the x, y coordinate of the position and head direction of the agent (will be further developed)
        head_direction: ndarray
            Contains the x and y Coordinates of the position
        position: ndarray
            Contains the x and y Coordinates of the position
    self.history: dict
        Saved history over simulation steps (action, state, new_state, reward, global_steps)
    global_steps: int
        Counter of the number of steps in the environment
    room_width: int
        Size of the environment in the x direction
    room_depth: int
        Size of the environment in the y direction
    metadata: dict
        Dictionary containing the metadata of the children experiment
            doi: str
                Add the reference to the experiemental results
    observation: ndarray
        Fully observable environment, make_observation returns the state
        Array of the observation of the agent in the environment (Could be modified as the environments are evolves)
    action: ndarray (env_dim,env_dim)
        Array containing the action of the agent
        In this case the delta_x and detla_y increment to the respective coordinate x and y of the position
    reward: int
        The reward that the animal recieves in this state

    """

    def __init__(
        self,
        use_behavioral_data: bool = False,
        data_path: str = None,
        recording_index: int = None,
        environment_name: str = "Sargolini2006",
        verbose: bool = False,
        experiment_class=Sargolini2006Data,
        **env_kwargs,
    ):
        super().__init__(
            use_behavioral_data,
            data_path,
            recording_index,
            environment_name,
            verbose,
            experiment_class,
            **env_kwargs,
        )


class BasicSargolini2006(Hafting2008):
    def __init__(
        self,
        use_behavioral_data: bool = False,
        data_path: str = None,
        recording_index: int = None,
        environment_name: str = "Sargolini2006",
        verbose: bool = False,
        experiment_class=SargoliniDataTrajectory,
        **env_kwargs,
    ):
        super().__init__(
            use_behavioral_data,
            data_path,
            recording_index,
            environment_name,
            verbose,
            experiment_class,
            **env_kwargs,
        )

    def set_animal_data(
        self,
        recording_index: int = 0,
        tolerance: float = 1e-10,
        keep_history: bool = True,
    ):
        """No recording index to use in this particular dataset"""
        if keep_history:
            prev_hist = copy.copy(self.history)
            self.reset()
            self.history = prev_hist
        else:
            self.reset()

    def show_data(self, full_dataframe: bool = False):
        print("no dataframe with sessions, just pre-processed positions of all trajectories")

    def plot_recording_tetr(self, **kwargs):
        print("No tetrode data available")

    def plot_recorded_trajectory(self, **kwargs):
        NotImplemented



import copy
from typing import Union

import matplotlib as mpl
import numpy as np

from neuralplayground.arenas import Simple2D
from neuralplayground.experiments import Hafting2008Data


class Hafting2008(Simple2D):
    """Arena resembling Hafting2008 experimental setting

    Methods
    ----------
    __init__
       Initialise the class
    reset(self):
      Reset the environment variables
    step(self, action):
       Increment the global step count of the agent in the environment and updates the position of the agent according
       to the recordings of the specific chosen session

    Attribute (In addition to the ones in Simple2D class)
    ---------
    use_behavioral_data: bool
        If True, then uses the animal trajectories recorded in Hafting 2008
    experiment: neuralplayground.experiments.Hafting2008Data
        Experiment class object with neural recordings and animal trajectories
    """

    def __init__(
        self,
        use_behavioral_data: bool = False,
        data_path: str = None,
        recording_index: int = None,
        environment_name: str = "Hafting2008",
        verbose: bool = False,
        experiment_class=Hafting2008Data,
        **env_kwargs,
    ):
        """Initialise the class

        Parameters
        ----------
        use_behavioral_data: bool
            If True, then uses the animal trajectories recorded in Hafting 2008
        data_path: str
            if None, fetch the data from the NeuralPlayground data repository,
            else load data from given path
        recording_index: int
            if None, load data from default recording index of corresponding experiment class
        environment_name: str
            Name of the specific instantiation of the Hafting2008 class
        verbose: bool
            Set to True to show the information of the class
        experiment_class:
            Experiment class to be initialized
        env_kwargs: dict
            Leave empty in this class, the arena parameters and sampling values are set to resemble the experiment
            For full control over these parameters use Simple2D class
        """

        self.data_path = data_path
        self.environment_name = environment_name
        self.use_behavioral_data = use_behavioral_data
        self.experiment = experiment_class(
            data_path=self.data_path,
            experiment_name=self.environment_name,
            verbose=verbose,
            recording_index=recording_index,
        )
        self.arena_limits = self.experiment.arena_limits
        self.arena_x_limits, self.arena_y_limits = (
            self.arena_limits[0, :],
            self.arena_limits[1, :],
        )
        env_kwargs["arena_x_limits"] = self.arena_x_limits
        env_kwargs["arena_y_limits"] = self.arena_y_limits
        env_kwargs["agent_step_size"] = 1.0
        env_kwargs["time_step_size"] = 1 / 50  # Taken from experiment, 50 Hz movement sampling
        super().__init__(environment_name, **env_kwargs)
        if self.use_behavioral_data:
            self.state_dims_labels = [
                "x_pos",
                "y_pos",
                "head_direction_x",
                "head_direction_y",
            ]

    def reset(self, random_state: bool = False, custom_state: np.ndarray = None):
        """Reset the environment variables. If using behavioral data, it will reset the position to the
        initial position of the trajectory recorded in the experiment

        Parameters
        ----------
        random_state: bool
            If True, sample a new position uniformly within the arena, use default otherwise
        custom_state: np.ndarray
            If given, use this array to set the initial state

        Returns
        ----------
        observation: ndarray
            Because this is a fully observable environment, make_observation returns the state of the environment
            Array of the observation of the agent in the environment (Could be modified as the environments are evolves)

        self.state: ndarray (2,)
            Vector of the x and y coordinate of the position of the animal in the environment
        """
        # Reset to first position recorded in this session
        if self.use_behavioral_data:
            self.pos, self.head_dir = (
                self.experiment.position[0, :],
                self.experiment.head_direction[0, :],
            )
            custom_state = np.concatenate([self.pos, self.head_dir])
            return super().reset(random_state=False, custom_state=custom_state)
        # Default reset
        else:
            return super().reset(random_state=random_state, custom_state=custom_state)

    def set_animal_data(
        self,
        recording_index: int = 0,
        tolerance: float = 1e-10,
        keep_history: bool = True,
    ):
        """Set position and head direction to be used by the Arena Class,
        See neuralplayground.experiments classes"""

        self.experiment.set_animal_data(recording_index=recording_index, tolerance=tolerance)
        if keep_history:
            prev_hist = copy.copy(self.history)
            self.reset()
            self.history = prev_hist
        else:
            self.reset()

    def show_data(self, full_dataframe: bool = False):
        """Print of available data recorded in the experiment

        Parameters
        ----------
        full_dataframe: bool
            if True, it will show all available experiment, a small sample otherwise

        Returns
        -------
        recording_list: Pandas dataframe
            List of available experiment, columns with rat_id, recording session and recorded variables
        """
        self.experiment.show_data(full_dataframe=full_dataframe)

    def plot_recording_tetr(
        self,
        recording_index: Union[int, tuple, list] = None,
        save_path: Union[str, tuple, list] = None,
        ax: Union[mpl.axes.Axes, tuple, list] = None,
        tetrode_id: Union[str, tuple, list] = None,
        bin_size: float = 2.0,
    ):
        """Check plot_recording_tetrode method from neuralplayground.experiments.Hafting2008Data"""
        return self.experiment.plot_recording_tetr(recording_index, save_path, ax, tetrode_id, bin_size)

    def plot_recorded_trajectory(
        self,
        recording_index: Union[int, tuple, list] = None,
        save_path: Union[str, tuple, list] = None,
        ax: Union[mpl.axes.Axes, tuple, list] = None,
        plot_every: int = 20,
    ):
        """Check plot_trajectory method from neuralplayground.experiments.Hafting2008Data"""
        return self.experiment.plot_trajectory(
            recording_index=recording_index,
            save_path=save_path,
            ax=ax,
            plot_every=plot_every,
        )

    def step(self, action: np.ndarray, normalize_step: bool = False, skip_every: int = 10):
        """Runs the environment dynamics. Increasing global counters.
        Given some action, return observation, new state and reward.
        If using behavioral data, the action argument is ignored, and instead inferred from the recorded trajectory
        in the experiment.

        Parameters
        ----------
        action: ndarray (2,)
            Array containing the action of the agent, in this case the delta_x and detla_y increment to position
        normalize_step: bool
            If true, the action is normalized to have unit size, then scaled by the agent step size
        skip_every: int
            When using behavioral data, the next state will be the position and head direction
            "skip_every" recorded steps after the current one, essentially reducing the sampling rate

        Returns
        -------
        reward: float
            The reward that the animal receives in this state
        new_state: ndarray
            Update the state with the updated vector of coordinate x and y of position and head directions respectively
        observation: ndarray
            Array of the observation of the agent in the environment
        """
        if not self.use_behavioral_data:
            return super().step(action)

        # In this case, the action is ignored and computed from the step in behavioral data recorded from the experiment
        if self.global_steps * skip_every >= self.experiment.position.shape[0] - 1:
            self.global_steps = np.random.choice(np.arange(skip_every))
        # Time elapsed since last reset
        self.global_time = self.global_steps * self.time_step_size

        # New state as "skip every" steps after the current one in the recording
        new_state = (
            self.experiment.position[self.global_steps * skip_every, :],
            self.experiment.head_direction[self.global_steps * skip_every, :],
        )
        new_state = np.concatenate(new_state)

        # Inferring action from recording
        action = new_state - self.state
        reward = self.reward_function(action, state=self.state)
        transition = {
            "action": action,
            "state": self.state,
            "next_state": new_state,
            "reward": reward,
            "step": self.global_steps,
        }
        self.history.append(transition)
        self.state = new_state
        observation = self.make_observation()
        self._increase_global_step()
        return observation, new_state, reward



import numpy as np
import pytest
from tqdm import tqdm

from neuralplayground.agents import AgentCore, Stachenfeld2018, Weber2018
from neuralplayground.arenas import BasicSargolini2006


@pytest.fixture
def get_environment():
    env = BasicSargolini2006()
    return [
        env,
    ]


class Testmodelcore(object):
    @pytest.fixture
    def init_model(self, get_environment):
        agent = AgentCore()

        return [
            agent,
        ]

    def test_init_model(self, init_model):
        assert isinstance(init_model[0], AgentCore)


class TestWeber2018(Testmodelcore):
    @pytest.fixture
    def init_model(self, get_environment):
        exc_eta = 2e-4
        inh_eta = 8e-4
        model_name = "model_example"
        sigma_exc = np.array([0.05, 0.05])
        sigma_inh = np.array([0.1, 0.1])
        Ne = 10
        Ni = 10
        Nef = 1
        Nif = 1
        alpha_i = 1
        alpha_e = 1
        we_init = 1.0
        wi_init = 1.5
        agent_step_size = 0.1
        env = get_environment[0]

        agent = Weber2018(
            model_name=model_name,
            exc_eta=exc_eta,
            inh_eta=inh_eta,
            sigma_exc=sigma_exc,
            sigma_inh=sigma_inh,
            Ne=Ne,
            Ni=Ni,
            agent_step_size=agent_step_size,
            ro=1,
            Nef=Nef,
            Nif=Nif,
            room_width=env.room_width,
            room_depth=env.room_depth,
            alpha_i=alpha_i,
            alpha_e=alpha_e,
            we_init=we_init,
            wi_init=wi_init,
        )
        return [
            agent,
        ]

    def test_init_model(self, init_model):
        assert isinstance(init_model[0], Weber2018)

    def test_plot_rates(self, init_model):
        init_model[0].plot_rates()

    def test_agent_interaction(self, init_model, get_environment):
        env = get_environment[0]
        total_iters = 0
        n_steps = 1
        obs, state = env.reset()
        obs = obs[:2]
        # for i in tqdm(range(env.total_number_of_steps)):
        for i in tqdm(range(n_steps)):
            # Observe to choose an action
            action = init_model[0].act(obs)
            init_model[0].update()
            # rate = agent.update()
            # Run environment for given action
            obs, state, reward = env.step(action)
            obs = obs[:2]
            total_iters += 1


class TestStachenfeld2018(Testmodelcore):
    @pytest.fixture
    def init_model(self, get_environment):
        agent_step_size = 10
        discount = 0.9
        threshold = 1e-6
        lr_td = 1e-2
        t_episode = 50
        n_episode = 50
        state_density = 1 / agent_step_size
        env = get_environment[0]
        twoDvalue = True
        agent = Stachenfeld2018(
            discount=discount,
            t_episode=t_episode,
            n_episode=n_episode,
            threshold=threshold,
            lr_td=lr_td,
            room_width=env.room_width,
            room_depth=env.room_depth,
            state_density=state_density,
            twoD=twoDvalue,
        )
        return [
            agent,
        ]

    def test_agent_interaction(self, init_model, get_environment):
        env = get_environment[0]
        total_iters = 0
        n_steps = 1
        obs, state = env.reset()
        obs = obs[:2]
        # for i in tqdm(range(env.total_number_of_steps)):
        for i in tqdm(range(n_steps)):
            # Observe to choose an action
            action = init_model[0].act(obs)
            init_model[0].update()
            # rate = agent.update()
            # Run environment for given action
            obs, state, reward = env.step(action)
            obs = obs[:2]
            total_iters += 1

    def test_init_model(self, init_model):
        assert isinstance(init_model[0], Stachenfeld2018)

    def test_plot_sr_ground_truth(self, init_model):
        sr = init_model[0].update_successor_rep()  # Choose your type of Update
        init_model[0].plot_eigen(sr, eigen=(0,), save_path=None)
        init_model[0].plot_eigen(sr, eigen=(0, 1), save_path=None)

    def test_plot_sr_td(self, init_model):
        sr_td = init_model[0].update_successor_rep_td_full()  # Choose your type of Update
        init_model[0].plot_eigen(sr_td, eigen=(0,), save_path=None)

    def test_plot_sr_sum(self, init_model):
        sr_sum = init_model[0].successor_rep_sum()
        init_model[0].plot_eigen(sr_sum, eigen=(0,), save_path=None)



import os
import pickle

import numpy as np
import pandas as pd
from deepdiff import DeepDiff
from gymnasium import Env, spaces


class Environment(Env):
    """Abstract parent environment class
    Methods
    ----------
    __init__(self, environment_name: str = "Environment", **env_kwargs):
        Initialize the class. env_kwargs arguments are specific for each of the child environments and
        described in their respective class
    make_observation(self):
        Returns the current state of the environment
    step(self, action):
        Runs the environment dynamics. Given some action, return observation, new state and reward.
    reset(self):
        Re-initialize state and global counters. Returns observation and re-setted state
    save_environment(self, save_path: str):
        Save current variables of the object to re-instantiate the environment later
    restore_environment(self, save_path: str):
        Restore environment saved using save_environment method
    get_trajectory_data(self):
        Returns interaction history

    Attributes
    ----------
    state: array
        Empty array for this abstract class
    history: list
        list containing transition history
    time_step_size: float
        time step in second traverse when calling step method
    metadata: dict
        dictionary with extra metadata that might be available in other classes
    env_kwags: dict
        Arguments given to the init method
    global_steps: int
        number of calls to step method, set to 0 when calling reset
    global_time: float
        time simulating environment through step method, then global_time = time_step_size * global_steps
    observation_space: gym.spaces
        specify the range of observations as in openai gym
    action_space: gym.spaces
        specify the range of actions as in openai gym
    """

    def __init__(self, environment_name: str = "Environment", **env_kwargs):
        """Initialisation of Environment class

        Parameters
        ----------
        environment_name: str
            environment name for the specific instantiation of the object
        env_kwargs: dict
            Define within each subclass for specific environments
            time_step_size: float
               Size of the time step in seconds
        """
        self.environment_name = environment_name
        if "time_step_size" in env_kwargs.keys():
            self.time_step_size = env_kwargs["time_step_size"]  # All environments should have this (second units)
        else:
            self.time_step_size = 1.0
        self.env_kwargs = env_kwargs  # Parameters of the environment
        self.metadata = {"env_kwargs": env_kwargs}  # Define within each subclass for specific environments
        self.state = np.array([])
        self.observation_space = spaces.Box(low=-np.inf, high=np.inf, shape=(0,), dtype=np.float64)
        self.action_space = spaces.Box(low=-np.inf, high=np.inf, shape=(0,), dtype=np.float64)
        self.history = []
        # Initializing global counts
        self.global_steps = 0
        self.global_time = 0

    def make_observation(self):
        """Just take the state and returns an array of sensory information
        In more complex cases, the observation might be different from the internal state of the environment

        Returns
        -------
        self.state:
            Define within each subclass for specific environments
            Variable containing the state of the environment (eg. position in the environment)
        """
        return self.state

    def step(self, action=None):
        """Runs the environment dynamics. Increasing global counters.
        Given some action, return observation, new state and reward.

        Parameters
        ----------
        action:
            Abstract argument representing the action from the agent

        Returns
        -------
        observation:
            Define within each subclass for specific environments
            Any set of observation that is made by the agent in the environment (position, visual features,...)
        self.state:
            Define within each subclass for specific environments
            Variable containing the state of the environment (eg.position in the environment)
        reward: int
            The reward that the animal receives in this state transition
        """
        observation = self.make_observation()  # Build sensory info from current state
        reward = self.reward_function(action, self.state)
        self._increase_global_step()
        # state should be updated as well
        return observation, self.state, reward

    def _increase_global_step(self):
        self.global_steps += 1
        self.global_time += self.time_step_size

    def reset(self):
        """Re-initialize state. Returns observation and re-setted state

        Returns
        -------
        observation:
            Define within each subclass for specific environments.
            Any set of observation that is made by the agent in the environment (position, visual features,...)
        self.state:
            Define within each subclass for specific environments
            Variable containing the state of the environment (eg.position in the environment)
        """
        self.global_steps = 0
        self.global_time = 0
        observation = self.make_observation()
        return observation, self.state

    def save_environment(self, save_path: str):
        """Save current variables of the object to re-instantiate the environment later

        Parameters
        ----------
        save_path: str
            Path to save the environment
        """
        pickle.dump(self.__dict__, open(os.path.join(save_path), "wb"), pickle.HIGHEST_PROTOCOL)

    def restore_environment(self, save_path: str):
        """Restore environment saved using save_environment method

        Parameters
        ----------
        save_path: str
            Path to retrieve the environment
        """
        self.__dict__ = pd.read_pickle(save_path)

    def __eq__(self, other):
        """Check if two environments are equal by comparing all of its attributes

        Parameters:
        ----------
        other: Environment
            Another instantiation of the environment
        return: bool
            True if self and other are the same exact environment
        """
        diff = DeepDiff(self.__dict__, other.__dict__)
        if len(diff) == 0:
            return True
        else:
            return False

    def get_trajectory_data(self):
        """Returns interaction history"""
        return self.history

    def reward_function(self, action, state):
        """Code reward curriculum here as a function of action and state
        and attributes of the environment if you want

        Parameters
        ----------
        action:
            same as step method argument
        state:
            same as state attribute of the class

        Returns
        -------
        reward: float
            reward given the parameters
        """
        reward = 0
        return reward


if __name__ == "__main__":
    env = Environment(environment_name="test_environment", time_step_size=0.5, one_kwarg_argument=10)
    print(env.__dict__)
    print(env.__dir__())
